#!/bin/bash
# Start script for unattended configuration of a Raspberry Pi

get_parameters() {
	# first the default values...
	new_locale='en_GB.UTF-8'
	new_timezone='Europe/London'
	new_ssh_setting=0
	new_wifi_country=GB
	new_wifi_ssid="Our network"
	new_wifi_password="Secret"
	packages_to_install=''

	# ...then see if values can be read from a file
	# then remove that (may contain password)
	# but save parameters for the next script back to the file
	cfgfile='/boot/one-time-script.conf'
	if [[ -f $cfgfile ]]; then
		while IFS='= ' read -r lhs rhs; do
			if [[ ! $lhs =~ ^\ *# && -n $lhs ]]; then	# skip comment-/malformed lines
				rhs="${rhs%%\#*}"    # Del end-of-line comments
				rhs="${rhs%"${rhs##*[^[:blank:]]}"}"  # Del trailing spaces/tabs
				rhs="${rhs%\"}"     # Del opening double-quotes 
				rhs="${rhs#\"}"     # Del closing double-quotes 
				rhs="${rhs%\'}"     # Del opening single-quotes 
				rhs="${rhs#\'}"     # Del closing single-quotes 
				declare -g $lhs="$rhs"
			fi
		done < $cfgfile && log "Read parameters from $cfgfile";
		echo "packages_to_install=($packages_to_install)" >> $cfgfile;
	else
		log "Using default parameters";
	fi;
}

# 3. PI USER PROFILE SETUP
# doing this before OS config because until reboot, sudo is confused by a new hostname
user_profile() {
	chmod a+w $templog
	cd /tmp
	sudo -u pi /bin/bash <<-END
		echo -n "Unsetting executable-bits of hidden files: " >> $templog;
		find /home/pi -type f -name '.*' -exec chmod -x \{\} + && echo OK >> $templog || echo FAILED >> $templog;
		if [[ -f /home/pi/.ssh/authorized_keys ]]; then
			echo -n "Making authorized ssh keys private: " >> $templog;
			chmod 0600 /home/pi/.ssh/authorized_keys && chmod 0700 /home/pi/.ssh && echo OK >> $templog || echo FAILED >> $templog;
		fi;
	END
}

# 4. OPERATING SYSTEM CONFIGURATION
os_config() {
	log -n "Change timezone: "
	raspi-config nonint do_change_timezone "$new_timezone" && log OK || log FAILED;

	log -n "Set SSH to "  # 0 = on, 1 = off
	[[ $new_ssh_setting == 0 ]] && log -n "on: " || log -n "off: ";
	raspi-config nonint do_ssh $new_ssh_setting && log OK || log FAILED;

	log -n "Set WiFi country: "
	raspi-config nonint do_wifi_country $new_wifi_country && log OK || log FAILED;

	log -n "Set WiFi login: "
	raspi-config nonint do_wifi_ssid_passphrase "$new_wifi_ssid" "$new_wifi_password" && log OK || log FAILED;

	log -n "Avoid language setting problems when logged in through SSH: "
	sed -i 's/^AcceptEnv LANG LC_\*/#AcceptEnv LANG LC_*/' /etc/ssh/sshd_config && log OK || log FAILED;

	log -n "Change locale: "
	raspi-config nonint do_change_locale "$new_locale" && log OK || log FAILED;
}

# 1. INTERNAL SCRIPT BUSINESS
# logging of the script's run
logfile=configuration.log
templog=/dev/shm/$logfile
log() {
	echo "$@" >> $templog;
}
log "Unattended configuration by $0";
exec 2>>$templog;	# log all errors

# stop this service from running at boot again
log -n "Remove automatic running of config script: ";
systemctl disable one-time-script.service && log OK || log FAILED;

# prepare for the package installation script to run on the next boot
log -n "Set up automatic running of package installation script: ";
systemctl enable packages-script.service && log OK || log FAILED;
# make sure it runs only after time synchronization (to avoid apt update errors)
systemctl enable systemd-time-wait-sync && log OK || log FAILED;


chmod +x /home/pi/osd/osd
chmod +x /home/pi/osd/multi_switch.sh
chmod +x /opt/retropie/configs/all/emulationstation/scripts/screensaver-start/1_lcd_sleep.sh
chmod +x /opt/retropie/configs/all/emulationstation/scripts/screensaver-stop/1_lcd_wake.sh

dkms remove xpi_gamecon/1.0 --all
#dkms add xpi_gamecon/1.0
#dkms build xpi_gamecon/1.0
dkms install xpi_gamecon/1.0

sed -zi '/xpi-gamecon/!s/$/\nxpi-gamecon/' /etc/modules

systemctl enable xpi_gamecon_shutdown.service
systemctl enable xpi_gamecon_reboot.service


get_parameters

# partitioning commands fail on Raspbian Stretch (9) and earlier;
# use the built-in resizing script
log -n "Expansion of the root partition: "
raspi-config nonint do_expand_rootfs && log OK || log FAILED;

log $'\nPI USER PROFILE SETUP';
user_profile # before os_config because until reboot, sudo is confused by a new hostname

log $'\nOPERATING SYSTEM CONFIGURATION';
os_config

# Write the log to the boot partition 
date > /boot/$logfile
cat $templog >> /boot/$logfile

reboot
